from fastapi import Request, Response

from debug_toolbar.panels import Panel
from debug_toolbar.types import Stats


class RoutesPanel(Panel):
    title = "Routes"
    template = "panels/routes.html"

    async def generate_stats(self, request: Request, response: Response) -> Stats:
        return {
            "routes": request.app.routes,
            "endpoint": request["endpoint"],
        }



import typing as t

from fastapi import Request, Response, __version__

from debug_toolbar.panels import Panel
from debug_toolbar.types import Stats


class VersionsPanel(Panel):
    title = "Versions"
    template = "panels/versions.html"

    @property
    def nav_subtitle(self) -> str:
        return f"FastAPI {__version__}"

    @property
    def scripts(self) -> t.List[str]:
        scripts = super().scripts
        scripts.append(self.url_for("debug_toolbar.static", path="js/versions.js"))
        return scripts

    async def generate_stats(self, request: Request, response: Response) -> Stats:
        try:
            import pkg_resources
        except ImportError:
            packages = []
        else:
            packages = sorted(
                pkg_resources.working_set,
                key=lambda pkg: pkg.project_name.lower(),
            )
        return {"packages": packages}



import typing as t
from time import perf_counter

from fastapi import Request, Response
from fastapi.concurrency import AsyncExitStack
from fastapi.dependencies.utils import solve_dependencies
from sqlalchemy import event
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.engine.default import DefaultExecutionContext
from sqlalchemy.orm import Session

from debug_toolbar.panels.sql import SQLPanel


class SQLAlchemyPanel(SQLPanel):
    title = "SQLAlchemy"

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self.engines: t.Set[Engine] = set()

    def register(self, engine: Engine) -> None:
        event.listen(engine, "before_cursor_execute", self.before_execute)
        event.listen(engine, "after_cursor_execute", self.after_execute)

    def unregister(self, engine: Engine) -> None:
        event.remove(engine, "before_cursor_execute", self.before_execute)
        event.remove(engine, "after_cursor_execute", self.after_execute)

    def before_execute(
        self,
        conn: Connection,
        cursor: t.Any,
        statement: str,
        parameters: t.Union[t.Sequence, t.Dict],
        context: DefaultExecutionContext,
        executemany: bool,
    ) -> None:
        conn.info.setdefault("start_time", []).append(perf_counter())

    def after_execute(
        self,
        conn: Connection,
        cursor: t.Any,
        statement: str,
        parameters: t.Union[t.Sequence, t.Dict],
        context: DefaultExecutionContext,
        executemany: bool,
    ) -> None:
        query = {
            "duration": (perf_counter() - conn.info["start_time"].pop(-1)) * 1000,
            "sql": statement,
            "params": parameters,
            "is_select": context.invoked_statement.is_select,
        }
        self.add_query(str(conn.engine.url), query)

    async def add_engines(self, request: Request):
        route = request["route"]

        if hasattr(route, "dependant"):
            if "fastapi_astack" not in request:
                async with AsyncExitStack() as stack:
                    request.scope["fastapi_astack"] = stack

            solved_result = await solve_dependencies(
                request=request,
                dependant=route.dependant,
                dependency_overrides_provider=route.dependency_overrides_provider,
            )
            for value in solved_result[0].values():
                if isinstance(value, Session):
                    self.engines.add(value.get_bind())

    async def process_request(self, request: Request) -> Response:
        await self.add_engines(request)

        for engine in self.engines:
            self.register(engine)
        try:
            response = await super().process_request(request)
        finally:
            for engine in self.engines:
                self.unregister(engine)
        return response



import functools
import json
import re
import typing as t
from urllib import parse

from anyio import CapacityLimiter
from anyio.lowlevel import RunVar
from fastapi import APIRouter, HTTPException, Request, Response, status
from fastapi.staticfiles import StaticFiles
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.routing import NoMatchFound
from starlette.types import ASGIApp

from debug_toolbar.api import render_panel
from debug_toolbar.settings import DebugToolbarSettings
from debug_toolbar.toolbar import DebugToolbar
from debug_toolbar.utils import import_string, matched_route


def show_toolbar(request: Request, settings: DebugToolbarSettings) -> bool:
    if settings.ALLOWED_IPS is not None:
        remote_addr, _ = request["client"]
        return request.app.debug and remote_addr in settings.ALLOWED_IPS
    return request.app.debug


class DebugToolbarMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, **settings: t.Any) -> None:
        super().__init__(app)
        self.settings = DebugToolbarSettings(**settings)
        self.show_toolbar = import_string(self.settings.SHOW_TOOLBAR_CALLBACK)
        self.router: APIRouter = app  # type: ignore[assignment]

        while not isinstance(self.router, APIRouter):
            self.router = self.router.app
        try:
            self.router.url_path_for(name="debug_toolbar.render_panel")
        except NoMatchFound:
            self.init_toolbar()

    def init_toolbar(self) -> None:
        self.router.get(
            self.settings.API_URL,
            name="debug_toolbar.render_panel",
        )(self.require_show_toolbar(render_panel))

        self.router.mount(
            self.settings.STATIC_URL,
            StaticFiles(packages=[__package__]),
            name="debug_toolbar.static",
        )

        @self.router.on_event("startup")
        def set_default_thread_limiter():
            RunVar("_default_thread_limiter").set(CapacityLimiter(1))

    async def dispatch(
        self,
        request: Request,
        call_next: RequestResponseEndpoint,
    ) -> Response:
        request.scope["route"] = matched_route(request)

        if (
            not request["route"]
            or not self.show_toolbar(request, self.settings)
            or self.settings.API_URL in request.url.path
        ):
            return await call_next(request)

        toolbar = DebugToolbar(request, call_next, self.settings)
        response = await toolbar.process_request(request)
        content_type = response.headers.get("Content-Type", "")
        is_html = content_type.startswith("text/html")

        if not (is_html or content_type == "application/json") or (
            "gzip" in response.headers.get("Accept-Encoding", "")
        ):
            return response

        await toolbar.record_stats(response)
        await toolbar.record_server_timing(response)
        toolbar.generate_server_timing_header(response)

        if is_html:
            async for body in response.body_iterator:  # type: ignore[attr-defined]
                if not isinstance(body, bytes):
                    body = body.encode(response.charset)

            decoded = body.decode(response.charset)
            pattern = re.escape(self.settings.INSERT_BEFORE)
            bits = re.split(pattern, decoded, flags=re.IGNORECASE)

            if len(bits) > 1:
                bits[-2] += toolbar.render_toolbar()
                body = self.settings.INSERT_BEFORE.join(bits).encode(response.charset)
                response.headers["Content-Length"] = str(len(body))

            async def stream() -> t.AsyncGenerator[bytes, None]:
                yield body

            response.body_iterator = stream()  # type: ignore[attr-defined]
        else:
            data = parse.quote(json.dumps(toolbar.refresh()))
            response.set_cookie(key="dtRefresh", value=data)

        return response

    def require_show_toolbar(
        self,
        f: t.Callable[..., t.Any],
    ) -> t.Callable[[t.Any], t.Any]:
        @functools.wraps(f)
        def decorator(request: Request, *args: t.Any, **kwargs: t.Any) -> t.Any:
            if not self.show_toolbar(request, self.settings):
                raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)
            return f(request, *args, **kwargs)

        return decorator



import typing as t

from fastapi import Request, Response
from pyinstrument import Profiler
from starlette.concurrency import run_in_threadpool

from debug_toolbar.panels import Panel
from debug_toolbar.types import Stats
from debug_toolbar.utils import is_coroutine


class ProfilingPanel(Panel):
    title = "Profiling"
    template = "panels/profiling.html"

    async def process_request(self, request: Request) -> Response:
        self.profiler = Profiler(**self.toolbar.settings.PROFILER_OPTIONS)
        is_async = is_coroutine(request["route"].endpoint)

        async def call(func: t.Callable) -> None:
            await run_in_threadpool(func) if not is_async else func()

        await call(self.profiler.start)

        try:
            response = await super().process_request(request)
        finally:
            await call(self.profiler.stop)
        return response

    async def generate_stats(self, request: Request, response: Response) -> Stats:
        return {"content": self.profiler.output_html()}



import typing as t
import uuid
from collections import OrderedDict

from anyio import create_task_group
from fastapi import Request, Response
from starlette.middleware.base import RequestResponseEndpoint

from debug_toolbar.panels import Panel
from debug_toolbar.settings import DebugToolbarSettings
from debug_toolbar.types import ServerTiming, Stats
from debug_toolbar.utils import get_name_from_obj, import_string

DT = t.TypeVar("DT", bound="DebugToolbar")

if t.TYPE_CHECKING:
    StoreDT = OrderedDict[str, DT]


class DebugToolbar:
    _store: "StoreDT" = OrderedDict()
    _panel_classes: t.Optional[t.Sequence[t.Type[Panel]]] = None

    def __init__(
        self,
        request: Request,
        call_next: RequestResponseEndpoint,
        settings: DebugToolbarSettings,
    ) -> None:
        self.request = request
        self.settings = settings
        panels = []

        for panel_class in self.get_panel_classes(
            settings.DEFAULT_PANELS + settings.PANELS,
        )[::-1]:
            panel = panel_class(self, call_next)
            panels.append(panel)

            if panel.enabled:
                call_next = panel.process_request

        self.process_request = call_next
        self._panels = {}

        while panels:
            panel = panels.pop()
            self._panels[panel.panel_id] = panel

        self.stats: Stats = {}
        self.server_timing_stats: t.Dict[str, ServerTiming] = {}
        self.store_id: t.Optional[str] = None

    @classmethod
    def get_panel_classes(
        cls: t.Type[DT],
        panels: t.Sequence[str],
    ) -> t.Sequence[t.Type[Panel]]:

        if cls._panel_classes is None:
            cls._panel_classes = [import_string(panel_path) for panel_path in panels]
        return cls._panel_classes

    @property
    def panels(self) -> t.Sequence[Panel]:
        return list(self._panels.values())

    def get_panel_by_id(self, panel_id: str) -> Panel:
        return self._panels[panel_id]

    @property
    def enabled_panels(self) -> t.Sequence[Panel]:
        return [panel for panel in self._panels.values() if panel.enabled]

    def store(self) -> None:
        if self.store_id:
            return

        self.store_id = uuid.uuid4().hex
        self._store[self.store_id] = self

        for _ in range(self.settings.RESULTS_CACHE_SIZE, len(self._store)):
            self._store.popitem(last=False)

    @classmethod
    def fetch(cls: t.Type[DT], store_id: str) -> t.Optional[DT]:
        return cls._store.get(store_id)

    async def record_stats(self, response: Response) -> None:
        async with create_task_group() as tg:
            for panel in self.enabled_panels:
                tg.start_soon(panel.record_stats, self.request, response)

    async def record_server_timing(self, response: Response) -> None:
        async with create_task_group() as tg:
            for panel in self.enabled_panels:
                tg.start_soon(panel.record_server_timing, self.request, response)

    def refresh(self) -> t.Dict[str, t.Any]:
        self.store()
        return {
            "storeId": self.store_id,
            "panels": {
                panel.panel_id: panel.nav_subtitle
                for panel in self.enabled_panels
                if panel.nav_subtitle
            },
        }

    def generate_server_timing_header(self, response: Response) -> None:
        data = []

        for panel in self.enabled_panels:
            stats = panel.get_server_timing_stats()
            if not stats:
                continue
            for key, title, value in stats:
                data.append(f'{panel.panel_id}_{key};dur={value};desc="{title}"')
        if data:
            response.headers["Server-Timing"] = ", ".join(data)

    def render(self, template: str, **context: t.Any) -> str:
        return self.settings.JINJA_ENV.get_template(template).render(
            toolbar=self,
            url_for=self.request.url_for,
            get_name_from_obj=get_name_from_obj,
            to_set=set,
            **context,
        )

    def render_toolbar(self) -> str:
        self.store()
        return self.render("base.html")



import html
import json
import re
import typing as t
from collections import defaultdict

import sqlparse
from fastapi import Request, Response
from fastapi.encoders import jsonable_encoder
from pydantic.color import Color
from sqlparse import tokens as T

from debug_toolbar.panels import Panel
from debug_toolbar.types import ServerTiming, Stats
from debug_toolbar.utils import color_generator

__all__ = ["SQLPanel", "parse_sql", "raw_sql"]

Stream = t.Generator[t.Tuple[T._TokenType, str], None, None]


class BoldKeywordFilter:
    def process(self, stream: Stream) -> Stream:
        for token_type, value in stream:
            is_keyword = token_type in T.Keyword
            if is_keyword:
                yield T.Text, "<strong>"
            yield token_type, html.escape(value)
            if is_keyword:
                yield T.Text, "</strong>"


class RawFilter:
    def process(self, stream: Stream) -> Stream:
        for token_type, value in stream:
            if token_type not in T.Number and token_type != T.String.Single:
                yield token_type, value


def simplify(sql: str) -> str:
    expr = r"SELECT</strong> (...........*?) <strong>FROM"
    return re.sub(expr, "SELECT</strong> &#8226;&#8226;&#8226; <strong>FROM", sql)


class FilterStack(sqlparse.engine.FilterStack):
    def run(self, sql: str) -> str:
        self.postprocess.append(sqlparse.filters.SerializerUnicode())
        return "".join(super().run(sql))


def parse_sql(sql: str, aligned_indent: bool = False) -> str:
    stack = FilterStack()

    if aligned_indent:
        stack.enable_grouping()
        stack.stmtprocess.append(
            sqlparse.filters.AlignedIndentFilter(char="&nbsp;", n="<br/>")
        )
    stack.preprocess.append(BoldKeywordFilter())
    return stack.run(sql)


def raw_sql(sql: str) -> str:
    stack = FilterStack()
    stack.preprocess.append(RawFilter())
    return stack.run(sql)


def nqueries(n: int) -> str:
    return f"{n} {'query' if n == 1 else 'queries'}"


class SQLPanel(Panel):
    template = "panels/sql.html"

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._sql_time: float = 0
        self._queries: t.List[t.Tuple[str, t.Dict[str, t.Any]]] = []
        self._databases: t.Dict[str, t.Dict[str, t.Any]] = {}
        self._colors: t.Generator[Color, None, None] = color_generator()

    @property
    def nav_subtitle(self) -> str:
        return f"{nqueries(len(self._queries))} in {self._sql_time:.2f}ms"

    def add_query(self, alias: str, query: t.Dict[str, t.Any]) -> None:
        duration = query["duration"]
        sql = query["sql"]

        query.update(
            {
                "sql_formatted": parse_sql(sql, aligned_indent=True),
                "sql_simple": simplify(parse_sql(sql, aligned_indent=False)),
                "is_slow": duration > self.toolbar.settings.SQL_WARNING_THRESHOLD,
            }
        )
        if alias not in self._databases:
            self._databases[alias] = {
                "time_spent": duration,
                "num_queries": 1,
                "rgb_color": next(self._colors),
            }
        else:
            self._databases[alias]["time_spent"] += duration
            self._databases[alias]["num_queries"] += 1

        self._sql_time += duration
        self._queries.append((alias, jsonable_encoder(query)))

    async def generate_stats(self, request: Request, response: Response) -> Stats:
        trace_colors: t.Dict[t.Tuple[str, str], Color] = defaultdict(
            lambda: next(self._colors)
        )
        duplicates: t.Dict[str, t.Dict[t.Tuple[str, str], int]] = defaultdict(
            lambda: defaultdict(int)
        )
        similar: t.Dict[str, t.Dict[str, t.Any]] = defaultdict(lambda: defaultdict(int))
        width_ratio_tally = 0

        def dup_key(query: t.Dict[str, t.Any]) -> t.Tuple[str, str]:
            return (query["sql"], json.dumps(query["params"]))

        def sim_key(query: t.Dict[str, t.Any]) -> str:
            return query.get("raw", query.get("sql"))

        for alias, query in self._queries:
            duplicates[alias][dup_key(query)] += 1
            similar[alias][sim_key(query)] += 1
            try:
                width_ratio = (query["duration"] / self._sql_time) * 100
            except ZeroDivisionError:
                width_ratio = 0

            query.update(
                {
                    "trace_color": trace_colors[dup_key(query)],
                    "start_offset": width_ratio_tally,
                    "end_offset": width_ratio + width_ratio_tally,
                    "width_ratio": width_ratio,
                }
            )
            width_ratio_tally += width_ratio

        duplicates = {
            alias: {query: c for query, c in queries.items() if c > 1}
            for alias, queries in duplicates.items()
        }
        similar = {
            alias: {
                query: (c, next(self._colors)) for query, c in queries.items() if c > 1
            }
            for alias, queries in similar.items()
        }
        for alias, query in self._queries:
            try:
                query["sim_count"], query["sim_color"] = similar[alias][sim_key(query)]
                query["dup_count"] = duplicates[alias][dup_key(query)]
            except KeyError:
                continue

        for alias, info in self._databases.items():
            try:
                info["sim_count"] = sum(c for c, _ in similar[alias].values())
                info["dup_count"] = sum(c for c in duplicates[alias].values())
            except KeyError:
                continue

        return {
            "databases": self._databases,
            "queries": self._queries,
            "sql_time": self._sql_time,
        }

    async def generate_server_timing(
        self,
        request: Request,
        response: Response,
    ) -> ServerTiming:
        stats = self.get_stats()
        n = len(stats.get("queries", []))
        return [("sql", f"SQL {nqueries(n)}", stats.get("sql_time", 0))]



import typing as t

from fastapi import Request, Response, status

from debug_toolbar.panels import Panel
from debug_toolbar.responses import StreamingHTMLResponse


class RedirectsPanel(Panel):
    has_content = False
    nav_title = "Intercept redirects"
    template = "redirect.html"

    async def process_request(self, request: Request) -> Response:
        response = await super().process_request(request)

        if 300 <= response.status_code < 400:
            redirect_to = response.headers.get("Location")

            if redirect_to:

                async def content() -> t.AsyncGenerator[str, None]:
                    yield self.render(
                        redirect_to=redirect_to,
                        status_code=response.status_code,
                    )

                response = StreamingHTMLResponse(
                    content=content(),
                    status_code=status.HTTP_200_OK,
                )
        return response



import typing as t
from contextlib import contextmanager
from time import perf_counter

from fastapi import Request, Response
from tortoise.backends.base.client import BaseDBAsyncClient

try:
    from tortoise.connection import connections
except ImportError:
    connections = None  # type: ignore

from debug_toolbar.panels.sql import SQLPanel, raw_sql


class DBWrapper:
    def __init__(self, db: BaseDBAsyncClient, on_execute: t.Callable[..., t.Any]):
        self.db = db
        self.on_execute = on_execute

    def __getattr__(self, attr: str) -> t.Any:
        return getattr(self.db, attr)

    async def execute_insert(self, query: str, values: list) -> t.Any:
        with self.on_execute(self.db, query, values):
            return await self.db.execute_insert(query, values)

    async def execute_query(
        self,
        query: str,
        values: t.Optional[list] = None,
    ) -> t.Tuple[int, t.Sequence[dict]]:
        with self.on_execute(self.db, query, values):
            return await self.db.execute_query(query, values)

    async def execute_script(self, query: str) -> None:
        with self.on_execute(self.db, query):
            await self.db.execute_script(query)

    async def execute_many(self, query: str, values: t.List[list]) -> None:
        with self.on_execute(self.db, query, values):
            await self.db.execute_many(query, values)

    async def execute_query_dict(
        self,
        query: str,
        values: t.Optional[list] = None,
    ) -> t.List[dict]:
        with self.on_execute(self.db, query, values):
            return await self.db.execute_query_dict(query, values)


class TortoisePanel(SQLPanel):
    title = "Tortoise ORM"

    @contextmanager
    def on_execute(
        self,
        db: BaseDBAsyncClient,
        statement: str,
        values: t.Any = None,
    ) -> t.Iterator[None]:
        start_time = perf_counter()
        try:
            yield
        finally:
            query = {
                "duration": (perf_counter() - start_time) * 1000,
                "sql": statement,
                "raw": raw_sql(statement),
                "params": values,
                "is_select": statement.lower().strip().startswith("select"),
            }
            self.add_query(db.connection_name, query)

    async def process_request(self, request: Request) -> Response:
        assert connections is not None, "tortoise-orm >= 0.19.0 is required"

        for conn in connections.all():
            db = DBWrapper(conn, self.on_execute)
            connections.set(conn.connection_name, db)  # type: ignore[arg-type]
        try:
            response = await super().process_request(request)
        finally:
            for conn in connections.all():
                connections.set(conn.connection_name, conn.db)  # type: ignore
        return response



import inspect
import statistics
import typing as t
from time import perf_counter

from fastapi import Request, Response
from pydantic import BaseModel
from pydantic.fields import ModelField

from debug_toolbar.panels import Panel
from debug_toolbar.types import ServerTiming, Stats
from debug_toolbar.utils import pluralize

if t.TYPE_CHECKING:
    from pydantic.fields import LocStr, ValidateReturn
    from pydantic.types import ModelOrDc

_validate = ModelField.validate


class PydanticPanel(Panel):
    title = "Pydantic"
    template = "panels/pydantic.html"

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._validation_time: float = 0
        self._parent_ids: t.Dict[t.Any, str] = {}
        self._validations: t.Dict[str, t.Any] = {}

    @property
    def nav_subtitle(self) -> str:
        validation_count = len(self._validations)
        return (
            f"{validation_count} validation{pluralize(validation_count)}"
            f" in {self._validation_time:.2f}ms"
        )

    @property
    def validate(self) -> t.Callable[..., t.Any]:
        def decorator(
            field: ModelField,
            v: t.Any,
            values: t.Dict[str, t.Any],
            *,
            loc: "LocStr",
            cls: t.Optional["ModelOrDc"] = None,
        ) -> "ValidateReturn":
            parent_id = self._parent_ids.get(cls)
            field_id = str(id(field))

            if parent_id is not None:
                field_id = f"{parent_id}-{field_id}"
                self._validations[parent_id]["has_childs"] = True

            if field_id not in self._validations:
                self._validations[field_id] = {
                    "obj": field,
                    "ncalls": 1,
                    "durations": [],
                    "has_childs": False,
                    "class_name": cls.__name__ if cls else "",
                }
            else:
                self._validations[field_id]["ncalls"] += 1

            if (
                field.sub_fields is None
                and inspect.isclass(field.type_)
                and issubclass(field.type_, BaseModel)
            ):
                self._parent_ids[field.type_] = field_id

            start_time = perf_counter()
            result = _validate(field, v, values, loc=loc, cls=cls)
            duration = (perf_counter() - start_time) * 1000
            self._validations[field_id]["durations"].append(duration)

            if parent_id is None:
                self._validation_time += duration
            return result

        return decorator

    async def process_request(self, request: Request) -> Response:
        ModelField.validate = self.validate  # type: ignore[assignment]
        try:
            response = await super().process_request(request)
        finally:
            ModelField.validate = _validate  # type: ignore[assignment]
        return response

    async def generate_stats(self, request: Request, response: Response) -> Stats:
        return {
            "validation_time": self._validation_time,
            "validations": self._validations,
            "mean": statistics.mean,
        }

    async def generate_server_timing(
        self,
        request: Request,
        response: Response,
    ) -> ServerTiming:
        stats = self.get_stats()
        return [
            ("pydantic", "Pydantic validation time", stats.get("validation_time", 0)),
        ]



import datetime
import logging
import typing as t

from fastapi import Request, Response
from starlette.concurrency import run_in_threadpool

from debug_toolbar.panels import Panel
from debug_toolbar.types import Stats
from debug_toolbar.utils import is_coroutine, pluralize

try:
    import threading
except ImportError:
    threading = None  # type: ignore[assignment]


class LogCollector:
    def __init__(self) -> None:
        if threading is None:
            raise NotImplementedError(
                "threading module is not available, "
                "this panel cannot be used without it"
            )
        self.collections: t.Dict[int, t.List[t.Dict[str, t.Any]]] = {}

    def get_collection(
        self,
        thread_id: t.Optional[int] = None,
    ) -> t.List[t.Dict[str, t.Any]]:
        if thread_id is None:
            thread_id = threading.get_ident()
        if thread_id not in self.collections:
            self.collections[thread_id] = []
        return self.collections[thread_id]

    def clear_collection(self, thread_id: t.Optional[int] = None) -> None:
        if thread_id is None:
            thread_id = threading.get_ident()
        if thread_id in self.collections:
            del self.collections[thread_id]

    def collect(
        self,
        item: t.Dict[str, t.Any],
        thread_id: t.Optional[int] = None,
    ) -> None:
        self.get_collection(thread_id).append(item)


class ThreadTrackingHandler(logging.Handler):
    def __init__(self, collector: LogCollector) -> None:
        logging.Handler.__init__(self)
        self.collector = collector

    def emit(self, record: logging.LogRecord) -> None:
        try:
            message = record.getMessage()
        except Exception:
            message = "[Could not get log message]"

        self.collector.collect(
            {
                "message": message,
                "time": datetime.datetime.fromtimestamp(record.created),
                "level": record.levelname,
                "file": record.pathname,
                "line": record.lineno,
                "channel": record.name,
            }
        )


collector = LogCollector()
logging_handler = ThreadTrackingHandler(collector)
logging.root.addHandler(logging_handler)


class LoggingPanel(Panel):
    nav_title = "Logging"
    title = "Log messages"
    template = "panels/logging.html"

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._records: t.Dict[t.Any, t.List[t.Dict[str, t.Any]]] = {}

    @property
    def nav_subtitle(self) -> str:
        stats = self.get_stats()
        record_count = len(stats["records"])
        return f"{record_count} message{pluralize(record_count)}"

    async def process_request(self, request: Request) -> Response:
        if is_coroutine(request["route"].endpoint):
            self.thread_id = threading.get_ident()
        else:
            self.thread_id = await run_in_threadpool(threading.get_ident)

        collector.clear_collection(thread_id=self.thread_id)
        return await super().process_request(request)

    async def generate_stats(self, request: Request, response: Response) -> Stats:
        records = collector.get_collection(thread_id=self.thread_id)
        self._records[self.thread_id] = records
        collector.clear_collection(thread_id=self.thread_id)
        return {"records": records}
